from typing import Iterable

from fireant.dataset.fields import Field
from .pandas import Pandas


class CSV(Pandas):
    def __init__(self, metric: Field, *metrics: Iterable[Field], group_pagination=False, **kwargs):
        super().__init__(metric, *metrics, **kwargs)
        self.group_pagination = group_pagination

    def transform(self, data_frame, slicer, dimensions, references):
        result_df = super(CSV, self).transform(data_frame, slicer, dimensions, references)
        result_df.columns.names = [None]
        return result_df.to_csv(na_rep='')



import numpy as np
import pandas as pd

from fireant.dataset.totals import get_totals_marker_for_dtype
from fireant.utils import (
    alias_selector,
    reduce_data_frame_levels,
)
from .fields import (
    DataType,
    Field,
)
from ..reference_helpers import reference_alias


def _extract_key_or_arg(data_frame, key):
    return data_frame[key] \
        if key in data_frame \
        else key


class Operation:
    """
    The `Operation` class represents an operation in the `Slicer` API.
    """

    def apply(self, data_frame, reference):
        raise NotImplementedError()

    @property
    def metrics(self):
        raise NotImplementedError()

    @property
    def operations(self):
        return []


class _BaseOperation(Operation):
    data_type = DataType.number

    def __init__(self,
                 alias,
                 label,
                 args,
                 prefix: str = None,
                 suffix: str = None,
                 thousands: str = None,
                 precision: int = None):
        self.alias = alias
        self.label = label
        self.args = args
        self.prefix = prefix
        self.suffix = suffix
        self.thousands = thousands
        self.precision = precision

    def apply(self, data_frame, reference):
        raise NotImplementedError()

    @property
    def metrics(self):
        return [metric
                for metric in self.args
                if isinstance(metric, Field)]

    @property
    def operations(self):
        return [op_and_children
                for operation in self.args
                if isinstance(operation, Operation)
                for op_and_children in [operation] + operation.operations]

    def _group_levels(self, index):
        """
        Get the index levels that need to be grouped. This is to avoid apply the cumulative function across separate
        dimensions. Only the first dimension should be accumulated across.

        :param index:
        :return:
        """
        return index.names[1:]

    def __repr__(self):
        return self.alias


class _Cumulative(_BaseOperation):
    def __init__(self, arg):
        super(_Cumulative, self).__init__(
              alias='{}({})'.format(self.__class__.__name__.lower(),
                                    getattr(arg, 'alias', arg)),
              label='{}({})'.format(self.__class__.__name__,
                                    getattr(arg, 'label', arg)),
              args=[arg],
              prefix=getattr(arg, 'prefix'),
              suffix=getattr(arg, 'suffix'),
              thousands=getattr(arg, 'thousands'),
              precision=getattr(arg, 'precision'),
        )

    def apply(self, data_frame, reference):
        raise NotImplementedError()


class CumSum(_Cumulative):
    def apply(self, data_frame, reference):
        arg, = self.args
        df_key = alias_selector(reference_alias(arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_key] \
                .groupby(level=levels) \
                .cumsum()

        return data_frame[df_key].cumsum()


class CumProd(_Cumulative):
    def apply(self, data_frame, reference):
        arg, = self.args
        df_key = alias_selector(reference_alias(arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_key] \
                .groupby(level=levels) \
                .cumprod()

        return data_frame[df_key].cumprod()


class CumMean(_Cumulative):
    @staticmethod
    def cummean(x):
        return x.cumsum() / np.arange(1, len(x) + 1)

    def apply(self, data_frame, reference):
        arg = self.args[0]
        df_key = alias_selector(reference_alias(arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_key] \
                .groupby(level=levels) \
                .apply(self.cummean)

        return self.cummean(data_frame[df_key])


class RollingOperation(_BaseOperation):
    def __init__(self, arg, window, min_periods=None):
        super(RollingOperation, self).__init__(
              alias='{}({},{})'.format(self.__class__.__name__.lower(),
                                       getattr(arg, 'alias', arg),
                                       window),
              label='{}({},{})'.format(self.__class__.__name__,
                                       getattr(arg, 'label', arg),
                                       window),
              args=[arg],
              prefix=getattr(arg, 'prefix'),
              suffix=getattr(arg, 'suffix'),
              thousands=getattr(arg, 'thousands'),
              precision=getattr(arg, 'precision'),
        )
        self.window = window
        self.min_periods = min_periods

    def _should_adjust(self, other_operations):
        # Need to figure out if this rolling operation is has the largest window, and if it's the first of multiple
        # rolling operations if there are more than one operation sharing the largest window.
        first_max_rolling = list(sorted(other_operations, key=lambda operation: operation.window))[0]

        return first_max_rolling is self

    def apply(self, data_frame, reference):
        raise NotImplementedError()


class RollingMean(RollingOperation):
    def rolling_mean(self, x):
        return x.rolling(self.window, self.min_periods).mean()

    def apply(self, data_frame, reference):
        arg, = self.args
        df_alias = alias_selector(reference_alias(arg, reference))

        if isinstance(data_frame.index, pd.MultiIndex):
            levels = self._group_levels(data_frame.index)

            return data_frame[df_alias] \
                .groupby(level=levels) \
                .apply(self.rolling_mean)

        return self.rolling_mean(data_frame[df_alias])


class Share(_BaseOperation):
    def __init__(self, metric: Field, over: Field = None, precision=2):
        super(Share, self).__init__(
              alias='share({},{})'.format(getattr(metric, 'alias', metric),
                                          getattr(over, 'alias', over), ),
              label='Share of {} over {}'.format(getattr(metric, 'label', metric),
                                                 getattr(over, 'label', over)),
              args=[metric, over],
              suffix='%',
              precision=precision,
        )

    @property
    def metrics(self):
        return [self.args[0]]

    @property
    def metric(self):
        return self.args[0]

    @property
    def over(self):
        return self.args[1]

    def apply(self, data_frame, reference):
        metric, over = self.args
        f_metric_alias = alias_selector(reference_alias(metric, reference))

        if over is None:
            df = data_frame[f_metric_alias]
            return 100 * df / df

        if not isinstance(data_frame.index, pd.MultiIndex):
            marker = get_totals_marker_for_dtype(data_frame.index.dtype)
            totals = data_frame.loc[marker, f_metric_alias]
            return 100 * data_frame[f_metric_alias] / totals

        f_over_alias = alias_selector(over.alias)
        idx = data_frame.index.names.index(f_over_alias)
        group_levels = data_frame.index.names[idx:]
        over_dim_value = get_totals_marker_for_dtype(data_frame.index.levels[idx].dtype)
        totals_alias = (slice(None),) * idx + (slice(over_dim_value, over_dim_value),)

        totals = reduce_data_frame_levels(data_frame.loc[totals_alias, f_metric_alias], group_levels)

        def apply_totals(group_df):
            if not isinstance(totals, pd.Series):
                return 100 * group_df / totals

            n_index_levels = len(totals.index.names)
            extra_level_names = group_df.index.names[n_index_levels:]
            group_df = group_df.reset_index(extra_level_names, drop=True)
            share = 100 * group_df / totals[group_df.index]
            return pd.Series(share.values, index=group_df.index)

        return data_frame[f_metric_alias] \
            .groupby(level=group_levels) \
            .apply(apply_totals) \
            .reorder_levels(order=data_frame.index.names) \
            .sort_index()



from pypika import NullValue


class Modifier:
    wrapped_key = 'wrapped'

    def __init__(self, wrapped):
        wrapped_key = super().__getattribute__('wrapped_key')
        setattr(self, wrapped_key, wrapped)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return super().__getattribute__(attr)

        wrapped_key = super().__getattribute__('wrapped_key')
        wrapped = super().__getattribute__(wrapped_key)
        return getattr(wrapped, attr)

    def __setattr__(self, attr, value):
        wrapped_key = super().__getattribute__('wrapped_key')
        if attr == wrapped_key:
            super().__setattr__(attr, value)
            return

        wrapped = super().__getattribute__(wrapped_key)

        if attr in wrapped.__dict__:
            setattr(wrapped, attr, value)
            return

        super().__setattr__(attr, value)

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        wrapped_key = super().__getattribute__('wrapped_key')
        wrapped = super().__getattribute__(wrapped_key)
        return '{}({})'.format(self.__class__.__name__, repr(wrapped))


class DimensionModifier(Modifier):
    wrapped_key = 'dimension'


class FilterModifier(Modifier):
    wrapped_key = 'filter'


class Rollup(DimensionModifier):
    @property
    def definition(self):
        return NullValue()


class OmitFromRollup(FilterModifier):
    pass



import copy
from functools import partial

from fireant.dataset.fields import Field
from fireant.dataset.intervals import (
    DatetimeInterval,
    week,
)
from fireant.dataset.references import YearOverYear
from pypika.terms import (
    ComplexCriterion,
    Criterion,
    Term,
)


def adapt_for_reference_query(reference_parts, database, dimensions, metrics, filters, references):
    if reference_parts is None:
        return database, dimensions, metrics, filters

    ref_dimension, time_unit, interval = reference_parts
    # Each group is guaranteed to have one reference and they will all be grouped together by reference type
    ref_database = _make_reference_database(database,
                                            ref_dimension,
                                            time_unit,
                                            interval)
    ref_metrics = _make_reference_metrics(metrics,
                                          references[0].reference_type.alias)
    offset_func = partial(database.date_add,
                          date_part=time_unit,
                          interval=interval)
    ref_dimensions = _make_reference_dimensions(dimensions,
                                                ref_dimension,
                                                offset_func)
    ref_filters = _make_reference_filters(filters,
                                          ref_dimension,
                                          offset_func)
    return ref_database, ref_dimensions, ref_metrics, ref_filters


def _make_reference_database(database, ref_dimension, time_unit, interval):
    # NOTE: In the case of weekly intervals with YoY references, the trunc date function needs to adjust for weekday
    # to keep things aligned. To do this, the date is first shifted forward a year before being truncated by week
    # and then shifted back.
    offset_for_weekday = isinstance(ref_dimension, DatetimeInterval) \
                         and ref_dimension.interval_key == week.keywords['interval_key'] \
                         and YearOverYear.time_unit == time_unit

    if offset_for_weekday:
        return _monkey_patch_align_weekdays(database, time_unit, interval)

    return database


def _make_reference_dimensions(dimensions, ref_dimension, offset_func):
    def replace_reference_dimension(dimension):
        ref_dimension = copy.copy(dimension)
        if hasattr(ref_dimension, 'dimension'):
            ref_dimension.dimension = copy.copy(dimension.dimension)
        ref_dimension.definition = offset_func(ref_dimension.definition)
        return ref_dimension

    return [replace_reference_dimension(dimension)
            if dimension is ref_dimension
            else dimension
            for dimension in dimensions]


def _make_reference_metrics(metrics, ref_key):
    return [Field(metric.alias + '_' + ref_key,
                  metric.definition,
                  label=metric.label,
                  prefix=metric.prefix,
                  suffix=metric.suffix,
                  precision=metric.precision)
            for metric in metrics]


def _make_reference_filters(filters, ref_dimension, offset_func):
    """
    Copies and replaces the reference dimension's definition in all of the filters applied to a slicer query.

    This is used to shift the dimension filters to fit the reference window.

    :param filters:
    :param ref_dimension:
    :param offset_func:
    :return:
    """
    offset_ref_dimension_definition = offset_func(ref_dimension.definition)

    reference_filters = []
    for ref_filter in map(copy.deepcopy, filters):
        ref_filter.definition = _apply_to_term_in_criterion(ref_dimension.definition,
                                                            offset_ref_dimension_definition,
                                                            ref_filter.definition)
        reference_filters.append(ref_filter)

    return reference_filters


def _monkey_patch_align_weekdays(database, time_unit, interval):
    original_trunc_date = database.__class__.trunc_date

    def trunc_date(definition, _):
        offset = original_trunc_date(database, definition, week.keywords['interval_key'])
        return database.date_add(offset, time_unit, -interval)

    # Copy the database to avoid side effects then monkey patch the trunc date function with the correction for weekday
    database = copy.deepcopy(database)
    database.trunc_date = trunc_date
    return database


def _apply_to_term_in_criterion(target: Term,
                                replacement: Term,
                                criterion: Criterion):
    """
    Finds and replaces a term within a criterion.  This is necessary for adapting filters used in reference queries
    where the reference dimension must be offset by some value.  The target term is found inside the criterion and
    replaced with the replacement.

    :param target:
        The target term to replace in the criterion. It will be replaced in all locations within the criterion with
        the func applied to itself.
    :param replacement:
        The replacement for the term.
    :param criterion:
        The criterion to replace the term in.
    :return:
        A criterion identical to the original criterion arg except with the target term replaced by the replacement arg.
    """
    if isinstance(criterion, ComplexCriterion):
        criterion.left = _apply_to_term_in_criterion(target, replacement, criterion.left)
        criterion.right = _apply_to_term_in_criterion(target, replacement, criterion.right)
        return criterion

    for attr in ['term', 'left', 'right']:
        if hasattr(criterion, attr) and str(getattr(criterion, attr)) == str(target):
            setattr(criterion, attr, replacement)

    return criterion



from pypika import (
    Query,
    enums,
    functions as fn,
    terms,
)


class Database(object):
    """
    This is a abstract base class used for interfacing with a database platform.
    """

    # The pypika query class to use for constructing queries
    query_cls = Query

    slow_query_log_min_seconds = 15

    def __init__(self, host=None, port=None, database=None, max_processes=2, max_result_set_size=200000,
                 cache_middleware=None):
        self.host = host
        self.port = port
        self.database = database
        self.max_processes = max_processes
        self.max_result_set_size = max_result_set_size
        self.cache_middleware = cache_middleware

    def connect(self):
        """
        This function must establish a connection to the database platform and return it.
        """
        raise NotImplementedError

    def trunc_date(self, field, interval):
        """
        This function must create a Pypika function which truncates a Date or DateTime object to a specific interval.
        """
        raise NotImplementedError

    def date_add(self, field: terms.Term, date_part: str, interval: int):
        """
        This function must add/subtract a Date or Date/Time object.
        """
        raise NotImplementedError

    def to_char(self, definition):
        return fn.Cast(definition, enums.SqlTypes.VARCHAR)



def reference_alias(metric, reference):
    """
    Format a metric key for a reference.

    :return:
        A string that is used as the key for a reference metric.
    """
    key = metric.alias

    if reference is None:
        return key

    return '{}_{}'.format(key, reference.alias)


def reference_label(metric, reference):
    """
    Format a metric label for a reference.

    :return:
        A string that is used as the display value for a reference metric.
    """
    label = metric.label or metric.alias

    if reference is None:
        return label

    return '{} {}'.format(label, reference.label)


def reference_prefix(metric, reference):
    """
    Return the prefix for a metric displayed for a reference (or no Reference)

    :return:
        A string that is used as the prefix for a reference metric.
    """
    if reference is not None and reference.delta_percent:
        return None
    return metric.prefix


def reference_suffix(metric, reference):
    """
    Return the suffix for a metric displayed for a reference (or no Reference)

    :return:
        A string that is used as the suffix for a reference metric.
    """
    if reference is not None and reference.delta_percent:
        return '%'
    return metric.suffix



from functools import (
    reduce,
    wraps,
)
from multiprocessing.pool import ThreadPool
from typing import (
    Iterable,
    Sized,
    Union,
)

import pandas as pd
import time

from fireant.database import Database
from fireant.dataset.fields import Field
from fireant.dataset.totals import get_totals_marker_for_dtype
from fireant.utils import (
    alias_selector,
    chunks,
)
from .finders import find_totals_dimensions
from .slow_query_logger import (
    query_logger,
    slow_query_logger,
)


def fetch_data(database: Database,
               queries: Union[Sized, Iterable],
               dimensions: Iterable[Field],
               share_dimensions: Iterable[Field] = (),
               reference_groups=()):
    iterable = [(str(query.limit(int(database.max_result_set_size))), database)
                for query in queries]

    with ThreadPool(processes=database.max_processes) as pool:
        results = pool.map(_exec, iterable)
        pool.close()

    return reduce_result_set(results, reference_groups, dimensions, share_dimensions)


def _exec(args):
    return _do_fetch_data(*args)


def db_cache(func):
    @wraps(func)
    def wrapper(query, database, *args):
        if database.cache_middleware is not None:
            return database.cache_middleware(func)(query, database, *args)
        return func(query, database, *args)

    return wrapper


def log(func):
    @wraps(func)
    def wrapper(query, database, *args):
        start_time = time.time()
        query_logger.debug(query)

        result = func(query, database, *args)

        duration = round(time.time() - start_time, 4)
        query_log_msg = '[{duration} seconds]: {query}'.format(duration=duration,
                                                               query=query)
        query_logger.info(query_log_msg)

        if database.slow_query_log_min_seconds is not None and duration >= database.slow_query_log_min_seconds:
            slow_query_logger.warning(query_log_msg)

        return result

    return wrapper


@db_cache
@log
def _do_fetch_data(query: str, database: Database):
    """
    Executes a query to fetch data from database middleware and builds/cleans the data as a data frame. The query
    execution is logged with its duration.

    :param database:
        instance of `fireant.Database`, database middleware
    :param query: Query string

    :return: `pd.DataFrame` constructed from the result of the query
    """
    with database.connect() as connection:
        return pd.read_sql(query, connection, coerce_float=True, parse_dates=True)


def reduce_result_set(results: Iterable[pd.DataFrame],
                      reference_groups,
                      dimensions: Iterable[Field],
                      share_dimensions: Iterable[Field]):
    """
    Reduces the result sets from individual queries into a single data frame. This effectively joins sets of references
    and concats the sets of totals.

    :param results: A list of data frame
    :param reference_groups: A list of groups of references (grouped by interval such as WoW, etc)
    :param dimensions: A list of dimensions, used for setting the index on the result data frame.
    :param share_dimensions: A list of dimensions from which the totals are used for calculating share operations.
    :return:
    """

    # One result group for each rolled up dimension. Groups contain one member plus one for each reference type used.
    result_groups = chunks(results, 1 + len(reference_groups))

    dimension_keys = [alias_selector(d.alias)
                      for d in dimensions]
    totals_dimension_keys = [alias_selector(d.alias)
                             for d in find_totals_dimensions(dimensions, share_dimensions)]
    dimension_dtypes = result_groups[0][0][dimension_keys].dtypes

    # Reduce each group to one data frame per rolled up dimension
    group_data_frames = []
    for i, result_group in enumerate(result_groups):
        if dimension_keys:
            result_group = [result.set_index(dimension_keys)
                            for result in result_group]

        base_df = result_group[0]
        reference_dfs = [_make_reference_data_frame(base_df, result, reference)
                         for result, reference_group in zip(result_group[1:], reference_groups)
                         for reference in reference_group]

        reduced = reduce(lambda left, right: pd.merge(left, right, how='outer', left_index=True, right_index=True),
                         [base_df] + reference_dfs)

        # If there are rolled up dimensions in this result set then replace the NaNs for that dimension value with a
        # marker to indicate totals.
        # The data frames will be ordered so that the first group will contain the data without any rolled up
        # dimensions, then followed by the groups with them, ordered by the last rollup dimension first.
        if totals_dimension_keys[:i]:
            reduced = _replace_nans_for_totals_values(reduced, dimension_dtypes)

        group_data_frames.append(reduced)

    return pd.concat(group_data_frames, sort=False) \
        .sort_index(na_position='first')


def _replace_nans_for_totals_values(data_frame, dtypes):
    # some things are just easier to do without an index. Reset it temporarily to replaxe NaN values with the rollup
    # marker values
    index_names = data_frame.index.names
    data_frame.reset_index(inplace=True)

    for dimension_key, dtype in dtypes.items():
        data_frame[dimension_key] = data_frame[dimension_key].fillna(get_totals_marker_for_dtype(dtype))

    return data_frame.set_index(index_names)


def _make_reference_data_frame(base_df, ref_df, reference):
    """
    This applies the reference metrics to the data frame given the base data frame and the reference data frame.

    When a reference is selected as a delta or a delta percentage, the calculation is performed here. Otherwise, the
    reference data frame is returned.

    :param base_df:
    :param ref_df:
    :param reference:
    :return:
    """
    mertric_column_indices = [i
                              for i, column in enumerate(ref_df.columns)
                              if column not in base_df.columns]
    ref_columns = [ref_df.columns[i] for i in mertric_column_indices]

    if not (reference.delta or reference.delta_percent):
        return ref_df[ref_columns]

    base_columns = [base_df.columns[i] for i in mertric_column_indices]

    # Select just the metric columns from the DF and rename them with the reference key as a suffix
    base_df, ref_df = base_df[base_columns].copy(), ref_df[ref_columns].copy()
    # Both data frame columns are renamed in order to perform the calculation below.
    base_df.columns = ref_df.columns = [column.replace(reference.reference_type.alias, reference.alias)
                                        for column in ref_columns]

    ref_delta_df = base_df - ref_df

    if reference.delta_percent:
        return 100. * ref_delta_df / ref_df
    return ref_delta_df



import re
from collections import OrderedDict

import pandas as pd

from fireant.dataset.fields import (
    DataType,
    Field,
)
from fireant.dataset.references import Reference
from fireant.dataset.totals import (
    DATE_TOTALS,
    NUMBER_TOTALS,
    TEXT_TOTALS,
    TOTALS_MARKERS,
)
from fireant.formats import (
    RAW_VALUE,
    TOTALS_VALUE,
    display_value,
    json_value,
    raw_value,
    return_none,
    safe_value,
)
from fireant.utils import (
    alias_for_alias_selector,
    alias_selector,
    setdeepattr,
    wrap_list,
)
from .pandas import Pandas
from ..reference_helpers import (
    reference_alias,
    reference_label,
    reference_prefix,
    reference_suffix,
)

TOTALS_LABEL = 'Totals'
METRICS_DIMENSION_ALIAS = 'metrics'
METRICS_DIMENSION = Field(METRICS_DIMENSION_ALIAS, None, data_type=DataType.text, label='')
F_METRICS_DIMENSION_ALIAS = alias_selector(METRICS_DIMENSION.alias)


def map_index_level(index, level, func):
    # If the index is empty, do not do anything
    if 0 == index.size:
        return index

    if isinstance(index, pd.MultiIndex):
        values = index.levels[level]
        return index.set_levels(values.map(func), level)

    assert level == 0

    return index.map(func)


class ReferenceItem:
    def __init__(self, item, reference):
        assert isinstance(reference, Reference)
        self.data_type = item.data_type
        self.alias = reference_alias(item, reference)
        self.label = reference_label(item, reference)
        self.prefix = reference_prefix(item, reference)
        self.suffix = reference_suffix(item, reference)
        self.thousands = item.thousands
        self.precision = item.precision


class TotalsItem:
    alias = TOTALS_VALUE
    label = TOTALS_LABEL
    prefix = suffix = precision = None


class ReactTable(Pandas):
    """
    This component does not work with react-table out of the box, some customization is needed in order to work with
    the transformed data.

    .. code-block:: jsx

        // A Custom TdComponent implementation is required by Fireant in order to render display values
        const TdComponent = ({
                               toggleSort,
                               className,
                               children,
                               ...rest
                             }) =>
            <div className={classNames('rt-td', className)} role="gridcell" {...rest}>
                {_.get(children, 'display', children.raw) || <span>&nbsp;</span>}
            </div>;

        const FireantReactTable = ({
                                config, // The payload from fireant
                              }) =>
            <ReactTable columns={config.columns}
                        data={config.data}
                        minRows={0}

                        TdComponent={ DashmoreTdComponent}
                        defaultSortMethod={(a, b, desc) => ReactTableDefaults.defaultSortMethod(a.raw, b.raw, desc)}>
            </ReactTable>;
    """

    def __init__(self, metric, *metrics: Field, pivot=(), transpose=False, sort=None, ascending=None,
                 max_columns=None):
        super(ReactTable, self).__init__(metric, *metrics,
                                         pivot=pivot,
                                         transpose=transpose,
                                         sort=sort,
                                         ascending=ascending,
                                         max_columns=max_columns)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ','.join(str(m) for m in self.items))

    @staticmethod
    def map_hyperlink_templates(df, dimensions):
        """
        Creates a mapping for each dimension to it's hyperlink template if it is possible to create the hyperlink
        template for it.

        The hyperlink template is a URL-like string containing curley braces enclosing dimension keys: `{dimension}`.
        While rendering this widget, the dimension key placeholders need to be replaced with the dimension values for
        that row.

        :param df:
            The result data set that is being transformed. The data frame SHOULD be pivoted/transposed if that step is
            required, before calling this function, in order to prevent the template from being included for the
            dimension if one of the required dimensions is pivoted.
        :param dimensions:
            The list of dimensions included in the query that created the result data set df.
        :return:
            A dict with the dimension key as the key and the hyperlink template as the value. Templates will only be
            included if it will be possible to fill in the required parameters.
        """
        hyperlink_templates = {}
        pattern = re.compile(r'{[^{}]+}')

        for dimension in dimensions:
            hyperlink_template = dimension.hyperlink_template
            if hyperlink_template is None:
                continue

            required_hyperlink_parameters = [alias_selector(argument[1:-1])
                                             for argument in pattern.findall(hyperlink_template)]

            # Check that all of the required dimensions are in the result data set. Only include the hyperlink template
            # in the return value of this function if all are present.
            unavailable_hyperlink_parameters = set(required_hyperlink_parameters) & set(df.index.names)
            if not unavailable_hyperlink_parameters:
                continue

            # replace the dimension keys with the formatted values. This will come in handy later when replacing the
            # actual values
            hyperlink_template = hyperlink_template.format(**{
                alias_for_alias_selector(argument): '{' + argument + '}'
                for argument in required_hyperlink_parameters
            })

            f_dimension_alias = alias_selector(dimension.alias)
            hyperlink_templates[f_dimension_alias] = hyperlink_template

        return hyperlink_templates

    @staticmethod
    def format_data_frame(data_frame):
        """
        This function prepares the raw data frame for transformation by formatting dates in the index and removing any
        remaining NaN/NaT values. It also names the column as metrics so that it can be treated like a dimension level.

        :param data_frame:
            The result set data frame
        :param dimensions:
        :return:
        """
        data_frame = data_frame.copy()
        data_frame.columns.name = F_METRICS_DIMENSION_ALIAS
        return data_frame

    @staticmethod
    def transform_index_column_headers(data_frame, field_map):
        """
        Convert the un-pivoted dimensions into ReactTable column header definitions.

        :param data_frame:
            The result set data frame
        :param field_map:
        :return:
            A list of column header definitions with the following structure.

        .. code-block:: jsx

            columns = [{
              Header: 'Column A',
              accessor: 'a',
            }, {
              Header: 'Column B',
              accessor: 'b',
            }]
        """
        columns = []
        if not isinstance(data_frame.index, pd.MultiIndex) and data_frame.index.name is None:
            return columns

        for f_dimension_alias in data_frame.index.names:
            dimension = field_map[f_dimension_alias]
            header = getattr(dimension, 'label', dimension.alias)

            columns.append({
                'Header': json_value(header),
                'accessor': safe_value(f_dimension_alias),
            })

        return columns

    @staticmethod
    def transform_data_column_headers(data_frame, field_map):
        """
        Convert the metrics into ReactTable column header definitions. This includes any pivoted dimensions, which will
        result in multiple rows of headers.

        :return:
        :param data_frame:
            The result set data frame
        :param field_map:
            A map to find metrics/operations based on their keys found in the data frame.
        :return:
            A list of column header definitions with the following structure.

        .. code-block:: jsx

            columns = [{
              Header: 'Column A',
              columns: [{
                Header: 'SubColumn A.0',
                accessor: 'a.0',
              }, {
                Header: 'SubColumn A.1',
                accessor: 'a.1',
              }]
            }, {
              Header: 'Column B',
              columns: [
                ...
              ]
            }]
        """

        def get_header(column_value, f_dimension_alias, is_totals):
            if f_dimension_alias == F_METRICS_DIMENSION_ALIAS or is_totals:
                item = field_map[column_value]
                return getattr(item, 'label', item.alias)

            if f_dimension_alias in field_map:
                field = field_map[f_dimension_alias]
                return display_value(column_value, field) or column_value

            return column_value

        def _make_columns(columns_frame, previous_level_values=()):
            """
            This function recursively creates the individual column definitions for React Table with the above tree
            structure depending on how many index levels there are in the columns.

            :param columns_frame:
                A data frame representing the columns of the result set data frame.
            :param previous_level_values:
                A tuple containing the higher level index level values used for building the data accessor path
            """
            f_dimension_alias = columns_frame.index.names[0]

            # Group the columns if they are multi-index so we can get the proper sub-column values. This will yield
            # one group per dimension value with the group data frame containing only the relevant sub-columns
            groups = columns_frame.groupby(level=0) \
                if isinstance(columns_frame.index, pd.MultiIndex) else \
                [(level, None) for level in columns_frame.index]

            columns = []
            for column_value, group in groups:
                is_totals = column_value in TOTALS_MARKERS | {TOTALS_LABEL}

                # All column definitions have a header
                column = {'Header': get_header(column_value, f_dimension_alias, is_totals)}

                level_values = previous_level_values + (column_value,)
                if group is not None:
                    # If there is a group, then drop this index level from the group data frame and recurse to build
                    # sub column definitions
                    next_level_df = group.reset_index(level=0, drop=True)
                    column['columns'] = _make_columns(next_level_df, level_values)

                else:
                    column['accessor'] = '.'.join(safe_value(value)
                                                  for value in level_values)

                if is_totals:
                    column['className'] = 'fireant-totals'

                columns.append(column)

            return columns

        # If the query only has a single metric, that level will be dropped, and set as data_frame.name
        dropped_metric_level_name = (data_frame.name,) if hasattr(data_frame, 'name') else ()

        return _make_columns(data_frame.columns.to_frame(),
                             dropped_metric_level_name)

    @staticmethod
    def transform_row_index(index_values, field_map, dimension_hyperlink_templates):
        # Add the index to the row
        row = {}
        for key, value in index_values.items():
            if key is None:
                continue

            field_alias = key
            field = METRICS_DIMENSION \
                if field_alias == METRICS_DIMENSION_ALIAS \
                else field_map[field_alias]

            data = {RAW_VALUE: raw_value(value, field)}
            display = display_value(value, field)
            if display is not None:
                data['display'] = display

            # If the dimension has a hyperlink template, then apply the template by formatting it with the dimension
            # values for this row. The values contained in `index_values` will always contain all of the required values
            # at this point, otherwise the hyperlink template will not be included.
            if key in dimension_hyperlink_templates:
                data['hyperlink'] = dimension_hyperlink_templates[key].format(**index_values)

            safe_key = safe_value(key)
            row[safe_key] = data

        return row

    @staticmethod
    def transform_row_values(series, fields, is_transposed):
        # Add the values to the row
        row = {}
        for key, value in series.iteritems():
            key = wrap_list(key)

            # Get the field for the metric
            metric_alias = (wrap_list(series.name)[0]
                            if is_transposed
                            else key[0])
            field = fields[metric_alias]

            data = {RAW_VALUE: raw_value(value, field)}
            display = display_value(value, field, date_as=return_none)
            if display is not None:
                data['display'] = display

            accessor_fields = [fields[field_alias]
                               for field_alias in series.index.names]
            accessor = [safe_value(value)
                        for value, field in zip(key, accessor_fields)]
            setdeepattr(row, accessor, data)

        return row

    @staticmethod
    def transform_data(data_frame, field_map, dimension_hyperlink_templates, is_transposed):
        """
        Builds a list of dicts containing the data for ReactTable. This aligns with the accessors set by
        #transform_dimension_column_headers and #transform_metric_column_headers

        :param data_frame:
            The result set data frame
        :param field_map:
            A mapping to all the fields in the slicer used for this query.
        :param dimension_hyperlink_templates:
        :param is_transposed:
        """
        index_names = data_frame.index.names

        def _get_field_label(alias):
            if alias not in field_map:
                return alias

            field = field_map[alias]
            return getattr(field, 'label', field.alias)

        # If the metric column was dropped due to only having a single metric, add it back here so the
        # formatting can be applied.
        if hasattr(data_frame, 'name'):
            metric_alias = data_frame.name
            data_frame = data_frame.copy()
            data_frame.columns = pd.MultiIndex.from_product([[metric_alias], data_frame.columns],
                                                            names=[F_METRICS_DIMENSION_ALIAS, data_frame.columns.name])

        rows = []
        for index, series in data_frame.iterrows():
            index = wrap_list(index)

            # Get a list of values from the index. These can be metrics or dimensions so it checks in the item map if
            # there is a display value for the value
            index_values = [_get_field_label(value) for value in index] \
                if is_transposed \
                else index
            index_display_values = OrderedDict(zip(index_names, index_values))

            rows.append({
                **ReactTable.transform_row_index(index_display_values,
                                                 field_map,
                                                 dimension_hyperlink_templates),
                **ReactTable.transform_row_values(series,
                                                  field_map,
                                                  is_transposed),
            })

        return rows

    def transform(self, data_frame, slicer, dimensions, references):
        """
        Transforms a data frame into a format for ReactTable. This is an object containing attributes `columns` and
        `data` which align with the props in ReactTable with the same name.

        :param data_frame:
            The result set data frame
        :param slicer:
            The slicer that generated the data query
        :param dimensions:
            A list of dimensions that were selected in the data query
        :param references:
            A list of references that were selected in the data query
        :return:
            An dict containing attributes `columns` and `data` which align with the props in ReactTable with the same
            names.
        """
        metric_map = OrderedDict([
            (
                alias_selector(reference_alias(i, ref)),
                ReferenceItem(i, ref) if ref is not None else i
            )
            for i in self.items
            for ref in [None] + references])
        dimension_map = {alias_selector(dimension.alias): dimension
                         for dimension in dimensions}
        field_map = {
            **metric_map,
            **dimension_map,

            # Add an extra item to map the totals markers to it's label
            NUMBER_TOTALS: TotalsItem,
            TEXT_TOTALS: TotalsItem,
            DATE_TOTALS: TotalsItem,
            TOTALS_LABEL: TotalsItem,
            alias_selector(METRICS_DIMENSION_ALIAS): METRICS_DIMENSION
        }
        all_dimensions_pivoted = 0 < len(dimensions) == len(self.pivot)  # has at least 1 dim and all are pivoted

        metric_aliases = list(metric_map.keys())
        dimension_aliases = [alias_selector(dimension.alias)
                             for dimension in self.pivot]

        df = self.format_data_frame(data_frame[metric_aliases])
        df = self.pivot_data_frame(df, dimension_aliases, self.transpose)

        dimension_columns = self.transform_index_column_headers(df, field_map)
        metric_columns = self.transform_data_column_headers(df, field_map)
        data = self.transform_data(df,
                                   field_map,
                                   is_transposed=self.transpose ^ all_dimensions_pivoted,
                                   dimension_hyperlink_templates=self.map_hyperlink_templates(df, dimensions))

        return {
            'columns': dimension_columns + metric_columns,
            'data': data,
        }



from typing import Iterable

from fireant.database import Database
from fireant.dataset.fields import Field
from fireant.dataset.filters import Filter
from fireant.dataset.intervals import DatetimeInterval
from fireant.dataset.joins import Join
from fireant.dataset.modifiers import Rollup
from fireant.utils import (
    alias_selector,
    flatten,
)
from pypika import (
    Table,
    functions as fn,
)
from .finders import (
    find_and_group_references_for_dimensions,
    find_filters_for_totals,
    find_joins_for_tables,
    find_required_tables_to_join,
    find_totals_dimensions,
)
from .reference_helper import adapt_for_reference_query
from .special_cases import apply_special_cases


def adapt_for_totals_query(totals_dimension, dimensions, filters):
    """
    Adapt filters for totals query. This function will select filters for total dimensions depending on the
    apply_filter_to_totals values for the filters. A total dimension with value None indicates the base query for
    which all filters will be applied by default.

    :param totals_dimension:
    :param dimensions:
    :param filters:
    :return:
    """
    is_totals_query = totals_dimension is not None
    raw_dimensions = [dimension.dimension
                      if isinstance(dimension, Rollup)
                      else dimension
                      for dimension in dimensions]

    # Get an index to split the dimensions before and after the totals dimension

    if is_totals_query:
        index = [i
                 for i, dimension in enumerate(dimensions)
                 if dimension is totals_dimension][0]
        totals_dims = [Rollup(dimension)
                       if i >= index
                       else dimension
                       for i, dimension in enumerate(raw_dimensions)]
        totals_filters = find_filters_for_totals(filters)

    else:
        totals_dims = raw_dimensions
        totals_filters = filters

    return totals_dims, totals_filters


@apply_special_cases
def make_slicer_query_with_totals_and_references(database,
                                                 table,
                                                 joins,
                                                 dimensions,
                                                 metrics,
                                                 operations,
                                                 filters,
                                                 references,
                                                 orders,
                                                 share_dimensions=()):
    """
    :param database:
    :param table:
    :param joins:
    :param dimensions:
    :param metrics:
    :param operations:
    :param filters:
    :param references:
    :param orders:
    :param apply_filter_to_totals:
    :param share_dimensions:
    :return:
    """

    """
    The following two loops will run over the spread of the two sets including a NULL value in each set:
     - reference group (WoW, MoM, etc.)
     - dimension with roll up/totals enabled (totals dimension)

    This will result in at least one query where the reference group and totals dimension is NULL, which shall be
    called base query. The base query will ALWAYS be present, even if there are zero reference groups or totals
    dimensions.

    For a concrete example, check the test case in :
    ```
    fireant.tests.queries.test_build_dimensions.QueryBuilderDimensionTotalsTests
        #test_build_query_with_totals_cat_dimension_with_references
    ```
    """
    totals_dimensions = find_totals_dimensions(dimensions, share_dimensions)
    totals_dimensions_and_none = [None] + totals_dimensions[::-1]

    reference_groups = find_and_group_references_for_dimensions(references)
    reference_groups_and_none = [(None, None)] + list(reference_groups.items())

    queries = []
    for totals_dimension in totals_dimensions_and_none:
        (query_dimensions,
         query_filters) = adapt_for_totals_query(totals_dimension,
                                                 dimensions,
                                                 filters)

        for reference_parts, references in reference_groups_and_none:
            (ref_database,
             ref_dimensions,
             ref_metrics,
             ref_filters) = adapt_for_reference_query(reference_parts,
                                                      database,
                                                      query_dimensions,
                                                      metrics,
                                                      query_filters,
                                                      references)
            query = make_slicer_query(ref_database,
                                      table,
                                      joins,
                                      ref_dimensions,
                                      ref_metrics,
                                      ref_filters,
                                      orders)

            # Add these to the query instance so when the data frames are joined together, the correct references and
            # totals can be applied when combining the separate result set from each query.
            query._totals = totals_dimension
            query._references = references

            queries.append(query)

    return queries


def make_slicer_query(database: Database,
                      base_table: Table,
                      joins: Iterable[Join] = (),
                      dimensions: Iterable[Field] = (),
                      metrics: Iterable[Field] = (),
                      filters: Iterable[Filter] = (),
                      orders: Iterable = ()):
    """
    Creates a pypika/SQL query from a list of slicer elements.

    This is the base implementation shared by two implementations: the query to fetch data for a slicer request and
    the query to fetch choices for dimensions.

    This function only handles dimensions (select+group by) and filtering (where/having), which is everything needed
    for the query to fetch choices for dimensions.

    The slicer query extends this with metrics, references, and totals.

    :param database:

    :param base_table:
        pypika.Table - The base table of the query, the one in the FROM clause
    :param joins:
        A collection of joins available in the slicer. This should include all slicer joins. Only joins required for
        the query will be used.
    :param dimensions:
        A collection of dimensions to use in the query.
    :param metrics:
        A collection of metircs to use in the query.
    :param filters:
        A collection of filters to apply to the query.
    :param orders:
        A collection of orders as tuples of the metric/dimension to order by and the direction to order in.

    :return:
    """
    query = database.query_cls.from_(base_table)
    elements = flatten([metrics, dimensions, filters])

    # Add joins
    join_tables_needed_for_query = find_required_tables_to_join(elements, base_table)
    for join in find_joins_for_tables(joins, base_table, join_tables_needed_for_query):
        query = query.join(join.table, how=join.join_type).on(join.criterion)

    # Add dimensions
    for dimension in dimensions:
        terms = make_terms_for_dimension(dimension, database.trunc_date)
        query = query.select(*terms)
        if not isinstance(dimension, Rollup):
            query = query.groupby(*terms)

    # Add filters
    for fltr in filters:
        query = query.having(fltr.definition) \
            if fltr.is_aggregate \
            else query.where(fltr.definition)

    # Add metrics
    term = make_terms_for_metrics(metrics)
    if term:
        query = query.select(*term)

    # In the case that the orders are determined by a field that is not selected as a metric or dimension, then it needs
    # to be added to the query.
    select_aliases = {el.alias for el in query._selects}
    for (term, orientation) in orders:
        query = query.orderby(term, order=orientation)

        if term.alias not in select_aliases:
            query = query.select(term)

    return query


def make_latest_query(database: Database,
                      base_table: Table,
                      joins: Iterable[Join] = (),
                      dimensions: Iterable[Field] = ()):
    query = database.query_cls.from_(base_table)

    # Add joins
    join_tables_needed_for_query = find_required_tables_to_join(dimensions, base_table)
    for join in find_joins_for_tables(joins, base_table, join_tables_needed_for_query):
        query = query.join(join.table, how=join.join_type).on(join.criterion)

    for dimension in dimensions:
        f_dimension_key = alias_selector(dimension.alias)
        query = query.select(fn.Max(dimension.definition).as_(f_dimension_key))

    return query


def make_terms_for_metrics(metrics):
    return [metric.definition.as_(alias_selector(metric.alias))
            for metric in metrics]


def make_terms_for_dimension(dimension, window=None):
    """
    Makes a list of pypika terms for a given slicer definition.

    :param dimension:
        A slicer dimension.
    :param window:
        A window function to apply to the dimension definition if it is a continuous dimension.
    :return:
        a list of terms required to select and group by in a SQL query given a slicer dimension. This list will contain
        either one or two elements. A second element will be included if the dimension has a definition for its display
        field.
    """
    f_alias = alias_selector(dimension.alias)

    if window and isinstance(dimension, DatetimeInterval):
        return [window(dimension.definition, dimension.interval_key).as_(f_alias)]

    return [dimension.definition.as_(f_alias)]


def make_orders_for_dimensions(dimensions):
    """
    Creates a list of ordering for a slicer query based on a list of dimensions. The dimensions's display definition is
    used preferably as the ordering term but the definition is used for dimensions that do not have a display
    definition.

    :param dimensions:
    :return:
        a list of tuple pairs like (term, orientation) for ordering a SQL query where the first element is the term
        to order by and the second is the orientation of the ordering, ASC or DESC.
    """

    # Use the same function to make the definition terms to force it to be consistent.
    # Always take the last element in order to prefer the display definition.
    definitions = [make_terms_for_dimension(dimension)[-1]
                   for dimension in dimensions]

    return [(definition, None)
            for definition in definitions]



from typing import Iterable

import pandas as pd

from fireant import formats
from fireant.dataset.fields import Field
from fireant.utils import (
    alias_selector,
    wrap_list,
)
from .base import (
    TransformableWidget,
)
from ..reference_helpers import (
    reference_alias,
    reference_label,
    reference_prefix,
    reference_suffix,
)

HARD_MAX_COLUMNS = 24


class Pandas(TransformableWidget):
    def __init__(self, metric: Field, *metrics: Iterable[Field],
                 pivot=(), transpose=False, sort=None, ascending=None, max_columns=None):
        super(Pandas, self).__init__(metric, *metrics)
        self.pivot = pivot
        self.transpose = transpose
        self.sort = sort
        self.ascending = ascending
        self.max_columns = min(max_columns, HARD_MAX_COLUMNS) \
            if max_columns is not None \
            else HARD_MAX_COLUMNS

    def transform(self, data_frame, slicer, dimensions, references):
        """
        WRITEME

        :param data_frame:
        :param slicer:
        :param dimensions:
        :param references:
        :return:
        """
        result = data_frame.copy()

        for metric in self.items:
            if any([metric.precision is not None,
                    metric.prefix is not None,
                    metric.suffix is not None]):
                df_alias = alias_selector(metric.alias)

                result[df_alias] = result[df_alias] \
                    .apply(lambda x: formats.display_value(x, metric))

            for reference in references:
                df_ref_alias = alias_selector(reference_alias(metric, reference))

                if reference.delta_percent:
                    result[df_ref_alias] = result[df_ref_alias].apply(lambda x: formats.display_value(
                          x,
                          reference_prefix(metric, reference),
                          reference_suffix(metric, reference),
                          metric.precision))

        if isinstance(data_frame.index, pd.MultiIndex):
            index_levels = [alias_selector(dimension.alias)
                            for dimension in dimensions]
            result = result.reorder_levels(index_levels)

        result = result[[alias_selector(reference_alias(item, reference))
                         for reference in [None] + references
                         for item in self.items]]

        if dimensions:
            result.index.names = [dimension.label or dimension.alias
                                  for dimension in dimensions]

        result.columns = pd.Index([reference_label(item, reference)
                                   for item in self.items
                                   for reference in [None] + references],
                                  name='Metrics')

        pivot_dimensions = [dimension.label or dimension.alias for dimension in self.pivot]
        return self.pivot_data_frame(result, pivot_dimensions, self.transpose)

    def pivot_data_frame(self, data_frame, pivot=(), transpose=False):
        """
        Pivot and transpose the data frame. Dimensions including in the `pivot` arg will be unshifted to columns. If
        `transpose` is True the data frame will be transposed. If there is only index level in the data frame (ie. one
        dimension), and that dimension is pivoted, then the data frame will just be transposed. If there is a single
        metric in the data frame and at least one dimension pivoted, the metrics column level will be dropped for
        simplicity.

        :param data_frame:
            The result set data frame
        :param pivot:
            A list of index aliases for `data_frame` of levels to shift
        :param transpose:
            A boolean true or false whether to transpose the data frame.
        :return:
            The shifted/transposed data frame
        """
        not_transforming_df = not (pivot or transpose)
        pivot_and_transpose_cancel_out = transpose and len(pivot) == len(data_frame.index.names)
        if not_transforming_df or pivot_and_transpose_cancel_out:
            return self.sort_data_frame(data_frame)

        # NOTE: Don't pivot a single dimension data frame. This turns the data frame into a series and pivots the
        # metrics anyway. Instead, transpose the data frame.
        should_transpose_instead_of_pivot = len(pivot) == len(data_frame.index.names)

        if pivot and not should_transpose_instead_of_pivot:
            data_frame = data_frame.unstack(level=pivot)

        if transpose or should_transpose_instead_of_pivot:
            data_frame = data_frame.transpose()

        # If there are more than one column levels and the last level is a single metric, drop the level
        if isinstance(data_frame.columns, pd.MultiIndex) and 1 == len(data_frame.columns.levels[0]):
            data_frame.name = data_frame.columns.levels[0][0]  # capture the name of the metrics column
            data_frame.columns = data_frame.columns.droplevel(0)  # drop the metrics level

        return self.sort_data_frame(data_frame)

    def sort_data_frame(self, data_frame):
        if not self.sort or len(data_frame) == 1:
            # If there are no sort arguments or the data frame is a single row, then no need to sort
            return data_frame

        # reset the index so all columns can be sorted together
        index_names = data_frame.index.names
        unsorted = data_frame.reset_index()
        column_names = list(unsorted.columns)

        ascending = self.ascending \
            if self.ascending is not None \
            else True

        sort = wrap_list(self.sort)
        sort_columns = [column_names[column]
                        for column in sort
                        if column < len(column_names)]

        if not sort_columns:
            return data_frame

        # ignore additional values for ascending if they do not align lengthwise with sort_columns
        # Default to the first value or None
        if isinstance(ascending, list) and len(ascending) != len(sort_columns):
            ascending = ascending[0] \
                if len(ascending) > 0 \
                else None

        return unsorted \
            .sort_values(sort_columns, ascending=ascending) \
            .set_index(index_names)
